use anyhow::{Context, Result};
use once_cell::sync::OnceCell;
use serde::{Deserialize, de::{Deserializer, MapAccess, Visitor}};
use std::{collections::HashMap, path::PathBuf, marker::PhantomData};
use std::fmt;

#[derive(Deserialize)]
pub struct Config {
    pub nickname: String,
    pub gateways: Vec<String>,
    pub api: String,
    pub memes: LowerCaseMap<String>,
}

static CONFIG: OnceCell<Config> = OnceCell::new();
pub static CONFIG_PATH: OnceCell<PathBuf> = OnceCell::new();

pub async fn try_get_config<'a>() -> Result<&'a Config> {
    match CONFIG.get() {
        None => {
            log::info!("Initializing config");
            let config = CONFIG_PATH.get_or_try_init::<_, anyhow::Error>(|| {
                let mut config = dirs::config_dir().context("Failed to get config dir")?;
                config.push("ruff/config.toml");
                Ok(config)
            })?;

            if let Some(p) = config.to_str() {
                log::info!("trying to read config at {}", p);
            }

            let bytes = tokio::fs::read(config).await?;
            let config = toml::from_slice::<Config>(&bytes)?;
            Ok(CONFIG.get_or_init(|| config))
        },
        Some(c) => Ok(c),
    }
}

pub struct LowerCaseMap<T> {
    pub map: HashMap<String, T>
}

impl<'de, T: Deserialize<'de>> Deserialize<'de> for LowerCaseMap<T> {
    fn deserialize<D>(deserializer: D) -> Result<LowerCaseMap<T>, D::Error>
    where
        D: Deserializer<'de>,
    {
        struct LowerCaseMapVisitor<T>(PhantomData<T>);

        impl<'de, T: Deserialize<'de>> Visitor<'de> for LowerCaseMapVisitor<T> {
            type Value = LowerCaseMap<T>;

            fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
                formatter.write_str("a key-value pair with a string key")
            }

            fn visit_map<A: MapAccess<'de>>(self, mut access: A) -> Result<Self::Value, A::Error> {
                let mut map = HashMap::new();
                while let Some((k, v)) = access.next_entry::<String, T>()? {
                    map.insert(k.to_lowercase(), v);
                }

                Ok(LowerCaseMap { map })
            }
        }

        deserializer.deserialize_map(LowerCaseMapVisitor(PhantomData))
    }
}

